#include	<iostream>
#include	<string.h>
#include	<stdlib.h>
#include	<new>
#include	"fsa.h"
#include	"nnode.h"

#define         START_CHAR      '^'

using namespace std;

struct signature {              /* dictionary file signature */
  char          sig[4];         /* automaton identifier (magic number) */
  char          ver;            /* automaton type number */
  char          filler;         /* char used as filler */
  char          annot_sep;      /* char that separates annotations from lex */
  char          gtl;            /* length of go_to field */
  char               type;
  char               version_major;
  unsigned short int version_minor;
  unsigned short int maxvysledku;
  unsigned short int maxvysledek;
  unsigned int       maxvysledky;
};
                      
/* Purpose:	Provide methods and common variables for building an automaton.
 * Methods:	automaton	- initiates roots of automaton and index;
 *		get_root	- returns root node of the automaton;
 *		build_fsa	- build the automaton;
 *		write_fsa	- writes the automaton to a file.
 * Variables:	root		- root of the automaton. */
class automaton {
private:
  node               *root;
public:
  char               FILLER;		/* character to be ignored (for fsa_guess) */
  char               type;
  char               version_major;
  unsigned short int version_minor;
  unsigned short int maxvysledek;
  unsigned short int maxvysledku;
  unsigned int       maxvysledky;
  automaton(void) { root = new node(); }
  node *get_root() const { return root; }
  node *set_root(node *new_root) { return (root = new_root); }
  int build_fsa(istream &infile);
  int write_fsa(ostream &out_file, const int make_numbers = FALSE);
};

const	int	WORD_BUFFER_LENGTH = 128;

int		fsa_arc_ptr::gtl = 2;	/* initialization (must be defined) */
int		fsa_arc_ptr::size = 4;	/* the same */

/* Purpose:	Writes the automaton in a binary form.
 * Parameters:	outfile		- (o) file to be written;
 *		make_numbers	- (i) whether to assign numbers to entries.
 * Returns:	TRUE if automaton written, FALSE otherwise.
 * Remarks:	Arcs must first be numbered in correct order.
 *		Arc number 0 represents the first arc of a node
 *		with no children :-). This makes it possible to use
 *		index 0 as a NULL pointer in a table of arcs.
 *
 *		The automaton file begins with the signature.
 *		The signature is followed by transitions of variable length. */
int automaton::write_fsa(ostream &outfile, const int make_numbers) {
  char		bytes[8];	/* output buffer for fsa arc */
  fsa_arc_ptr	*dummy;
  node		*meta_root;
  signature	sig_arc;
  int		result;
  int		gtl_calculated = FALSE;
  int		gtl;
  int		flag_bits = 3; // 3 bits from go_to taken by flags

  meta_root = new node;
  meta_root->add_child(START_CHAR, root);
  meta_root->hit_node();

  root->hit_node();		// to "register" it
  meta_root->number_arcs(0);	// only count arcs
  // we need to calcuate gtl first, as it is used in numbering arcs
  gtl = 1;
  for (int r = meta_root->no_of_arcs;
       ((r * (goto_offset+gtl) - meta_root->next_nodes * (gtl - 1))) >> ((8 * gtl) - flag_bits); ) gtl++;
  dummy->gtl = sig_arc.gtl = gtl;
  dummy->size = gtl + goto_offset;
  mark_inner(root, -1);		// pretend arcs were not numbered
				// otherwise number_arcs() would not work again
  root->no_of_arcs = goto_offset + gtl;
  gtl_calculated = TRUE;

  meta_root->number_arcs(gtl);	// assign addresses to nodes

  strncpy(sig_arc.sig, "\\fsa", 4);
  sig_arc.ver = 5;
  sig_arc.filler = '_';
  sig_arc.annot_sep = '+';
  long int arc_bytes = 0L;
  if (!gtl_calculated) {
    // calculate gtl (length of the go_to field)
    // initially, r is the highest value in the go_to field with flags cleared,
    // but counted as part of the go_to field
    sig_arc.gtl = 0;
    arc_bytes = meta_root->no_of_arcs * 8; // 3 bits from go_to taken by flags
    for (int r = arc_bytes; r; r >>= 8) sig_arc.gtl++;
    gtl = dummy->gtl = sig_arc.gtl;
    dummy->size = dummy->gtl + goto_offset;
  }
  sig_arc.type = type;
  sig_arc.maxvysledek = maxvysledek;
  sig_arc.maxvysledku = maxvysledku;
  sig_arc.maxvysledky = maxvysledky;
  sig_arc.version_major = version_major;
  sig_arc.version_minor = version_minor;
  if (!(outfile.write((char *)&sig_arc, sizeof sig_arc))) return FALSE;

  // write the sink node  
  for (int i = 0; i < dummy->size; i++) bytes[i] = 0;
  if (!(outfile.write(bytes, dummy->size))) return FALSE;

  // write the rest of the automaton
  result = meta_root->write_arcs(outfile);
  delete meta_root;
  return result;
}

void not_enough_memory(void) {
  cerr << "Not enough memory for the automaton\n";
  exit(4);
}

/* Purpose:	Make a copy of a string with more space allocated.
 * Parameters:	s		- (i/o) the string / its copy;
 *		allocated	- (i/o) size of old/new string;
 *		alloc_step	- (i) size of old string - size of new string.
 * Returns:	New string. */
char * grow_string(char *&s, int &allocated, const int alloc_step) {
  char *new_s;

  new_s = new char[allocated + alloc_step];
  memcpy(new_s, s, (size_t)allocated);
  allocated += alloc_step;
  delete [] s;
  s = new_s;
  return new_s;
}

/* Purpose:	Find how many letters of the word just read are already in
 *		the automaton. Find the node that contains the last letter.
 * Parameter:	start_node	- (i) where to start in the automaton;
 *		word		- (i) word whose common prefix is sought;
 *		common_prefix	- (o) information about the prefix;
 *		length		- (i) how many characters of word were
 *					already considered.
 * Returns:	common_prefix length.
 * Remarks:	I had to add `length' parameter, beacuse that was the only
 *		way to avoid using a global variable (I hate doing that).
 */
int find_common_prefix(node *start_node, const char *word, node **common_prefix, const int length) {
  arc_node	*offspr;

  for (int i = 0; i < start_node->get_no_of_kids(); i++) {
    offspr = start_node->get_children() + i;
    if (offspr->letter == *word) {
      if (offspr->child) return find_common_prefix(offspr->child, word + 1, common_prefix, length + 1);
      else break;
    }
  }
  *common_prefix = start_node;
  return length;
}

int automaton::build_fsa(istream &infile) {
  static char	*word_buffer;
  int		allocated = WORD_BUFFER_LENGTH;
  char		*word;
  char		*rest;
  node		*common_prefix;
  node		*last_node = NULL;
  char		junk;

  word_buffer = new char[WORD_BUFFER_LENGTH];
  word = &word_buffer[0];
  infile >> version_minor >> maxvysledek >> maxvysledku >> maxvysledky;
  infile.get(junk); // eat '\n';
  while (infile.get(word, allocated, '\n')) {
    infile.get(junk);				// eat '\n'
    while (junk != '\n') {
      // line was too long, allocate more space
      word = grow_string(word_buffer, allocated, WORD_BUFFER_LENGTH);
      word[allocated - WORD_BUFFER_LENGTH - 1] = junk;
      if (!(infile.get(word + allocated - WORD_BUFFER_LENGTH, WORD_BUFFER_LENGTH, '\n'))) {
	// This horrible code is provided as an attempt to counteract recent
	// absurd I/O library behaviour
	if (!(infile.bad()) && !(infile.eof())) {
	  // Last "junk" was probably the last character of the line.
	  // Were we to omit this, the next get() would fail!!!
	  infile.clear();
	  infile.get(junk);	// There should still be EOLN in the buffer!
	}
	break;
      }
      infile.get(junk);
    }
    // the rest is word - common_prefix
    rest = word + find_common_prefix(root, word, &common_prefix, 0);
    last_node = common_prefix;

    // We are retreating in the graph. This means that the last child
    // and possibly its children of the last node in prefix are
    // to be either registered (if there is no isomorphic subgraph
    // in the automaton, or deleted (i.e. replaced by pointers to
    // subgraphs already registered).
    if (last_node->fertile()) common_prefix->compress_or_register();

    // If the rest of word (after prefix) exists make a chain of nodes
    // with its consecutive letters, and attach it to the last node found in prefix.
    if (*rest) common_prefix->add_postfix(rest);
  }

  root->compress_or_register();
  return TRUE;
}

/* Returns:	0	- if OK;
 *		1	- if invalid options;
 *		2	- if automaton could not be built;
 *		3	- if automaton could not be written to a file;
 *		4	- if there is not enough memory to build the automaton. */
int main(const int argc, const char *argv[]) {
  char type = 0;
  
  if (argc >= 2) {
    for (int i = 1; i < argc; i++) {
      if (strcmp(argv[i], "-t") == 0) {
        if (++i < argc) {
          if      (strcmp(argv[i], "w-lt_is") == 0) type = 1;
          else if (strcmp(argv[i], "w_is")    == 0) type = 2;
          else if (strcmp(argv[i], "lt-w_is") == 0) type = 3;
          else if (strcmp(argv[i], "l-wt_is") == 0) type = 4;
          else if (strcmp(argv[i], "l-w_is")  == 0) type = 5;
          else if (strcmp(argv[i], "w-l_is")  == 0) type = 6;
          else if (strcmp(argv[i], "w-w_is")  == 0) type = 7;
          else if (strcmp(argv[i], "w-tw_is")  == 0) type = 8;
          else if (strcmp(argv[i], "wt-w_is")  == 0) type = 9;
          else if (strcmp(argv[i], "w-lt_fi") == 0) type = 1 + 128;
          else if (strcmp(argv[i], "w_fi")    == 0) type = 2 + 128;
          else if (strcmp(argv[i], "lt-w_fi") == 0) type = 3 + 128;
          else if (strcmp(argv[i], "l-wt_fi") == 0) type = 4 + 128;
          else if (strcmp(argv[i], "l-w_fi")  == 0) type = 5 + 128;
          else if (strcmp(argv[i], "w-l_fi")  == 0) type = 6 + 128;
          else if (strcmp(argv[i], "w-w_fi")  == 0) type = 7 + 128; 
          else if (strcmp(argv[i], "w-tw_fi")  == 0) type = 8 + 128; 
          else if (strcmp(argv[i], "wt-w_fi")  == 0) type = 9 + 128; 
          else {
            cerr << argv[0] << ": unknown type " << argv[i] << endl;
            return 1;
            }
          }
        else {
          cerr << argv[0] << ": -t without type" << endl;
          return 1;
          }
        }
      else {
	cerr << "bad usage" << endl;
	return 1;
      }
    }
  }

  set_new_handler(&not_enough_memory);

  automaton autom;
  autom.FILLER = '_';
  autom.type = type;
  autom.version_major = 1;
  if (!autom.build_fsa(cin)) {
    cerr << argv[0] << ": Could not build the automaton\n";
    return 2;
  }
  if (autom.write_fsa(cout)) return 0;
  cerr << argv[0] << ": Could not write the automaton\n";
  return 3;
}
